""":term:`BGP` session product block."""

from ipaddress import IPv4Address, IPv6Address

from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle
from pydantic_forms.types import strEnum


class IPFamily(strEnum):
    """Possible :term:`IP` families of a :term:`BGP` peering."""

    V4UNICAST = "v4unicast"
    V6UNICAST = "v6unicast"
    V4MULTICAST = "v4multicast"
    V6MULTICAST = "v6multicast"


class BGPSessionInactive(ProductBlockModel, lifecycle=[SubscriptionLifecycle.INITIAL], product_block_name="BGPSession"):
    """A :term:`BGP` session that is currently inactive. See :class:`BGPSession`."""

    peer_address: IPv4Address | IPv6Address | None = None
    bfd_enabled: bool | None = None
    bfd_interval: int | None = None
    bfd_multiplier: int | None = None
    families: list[IPFamily] | None = None
    has_custom_policies: bool | None = None
    authentication_key: str | None = None
    multipath_enabled: bool | None = None
    send_default_route: bool | None = None
    is_multi_hop: bool = False


class BGPSessionProvisioning(BGPSessionInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """A :term:`BGP` session that is currently being provisioned. See :class:`BGPSession`."""

    peer_address: IPv4Address | IPv6Address
    bfd_enabled: bool
    bfd_interval: int | None = None
    bfd_multiplier: int | None = None
    families: list[IPFamily]
    has_custom_policies: bool
    authentication_key: str
    multipath_enabled: bool
    send_default_route: bool
    is_multi_hop: bool


class BGPSession(BGPSessionProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """A :term:`BGP` session that is currently deployed in the network."""

    #: The peering address of the session.
    peer_address: IPv4Address | IPv6Address
    #: Whether :term:`BFD` is enabled.
    bfd_enabled: bool
    #: The :term:`BFD` interval, if enabled.
    bfd_interval: int | None = None
    #: The :term:`BFD` multiplier, if enabled.
    bfd_multiplier: int | None = None
    #: The list of :term:`IP` families enabled for this session.
    families: list[IPFamily]
    #: Whether any custom policies exist for this session.
    has_custom_policies: bool
    #: The authentication key of the :term:`BGP` session.
    authentication_key: str
    #: Whether multipath is enabled.
    multipath_enabled: bool
    #: Whether we send a last resort route.
    send_default_route: bool
    #: Whether this session is multi-hop or not. Defaults to no.
    is_multi_hop: bool
