"""Edge port product block.

Edge port sets the boundary between Geant network and an external entity that could also be a different technological
domain still managed by GEANT. In other words, an Edge port determines where the network ends.
"""

from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle, strEnum

from gso.products.product_blocks.router import RouterBlock, RouterBlockInactive, RouterBlockProvisioning
from gso.utils.types.interfaces import LAGMember, LAGMemberList, PhysicalPortCapacity


class EncapsulationType(strEnum):
    """Enum representing different Ethernet encapsulation service options.

    Null supports a single service on the port.
    Dot1Q supports multiple services for one customer or services for multiple customers.
    QinQ expands VLAN space by double-tagging frames.
    """

    DOT1Q = "dot1q"
    QINQ = "qinq"
    NULL = "null"


class EdgePortType(strEnum):
    """Types of edge ports."""

    CUSTOMER = "CUSTOMER"
    INFRASTRUCTURE = "INFRASTRUCTURE"
    PRIVATE = "PRIVATE"
    PUBLIC = "PUBLIC"
    RE_INTERCONNECT = "RE_INTERCONNECT"


class EdgePortBlockInactive(
    ProductBlockModel, lifecycle=[SubscriptionLifecycle.INITIAL], product_block_name="EdgePortBlock"
):
    """An edge port that's currently inactive. See :class:`EdgePortBlock`."""

    edge_port_node: RouterBlockInactive | None = None
    edge_port_name: str | None = None
    edge_port_description: str | None = None
    edge_port_enable_lacp: bool | None = None
    edge_port_encapsulation: EncapsulationType = EncapsulationType.DOT1Q
    edge_port_mac_address: str | None = None
    edge_port_member_speed: PhysicalPortCapacity | None = None
    edge_port_minimum_links: int | None = None
    edge_port_type: EdgePortType | None = None
    edge_port_ignore_if_down: bool = False
    edge_port_geant_ga_id: str | None = None
    edge_port_ae_members: LAGMemberList[LAGMember]


class EdgePortBlockProvisioning(EdgePortBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """An edge port that's being provisioned. See :class:`EdgePortBlock`."""

    edge_port_node: RouterBlockProvisioning
    edge_port_name: str
    edge_port_description: str | None = None
    edge_port_enable_lacp: bool
    edge_port_encapsulation: EncapsulationType = EncapsulationType.DOT1Q
    edge_port_mac_address: str | None = None
    edge_port_member_speed: PhysicalPortCapacity
    edge_port_minimum_links: int | None = None
    edge_port_type: EdgePortType
    edge_port_ignore_if_down: bool = False
    edge_port_geant_ga_id: str | None = None
    edge_port_ae_members: LAGMemberList[LAGMember]


class EdgePortBlock(EdgePortBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """An edge port that's currently deployed in the network."""

    #: The router that this edge port is connected to.
    edge_port_node: RouterBlock
    #: The name of the edge port, in our case, corresponds to the name of the :term:`LAG` interface.
    edge_port_name: str
    #: A description of the edge port.
    edge_port_description: str | None = None
    #: Indicates whether :term:`LACP` is enabled for this edge port.
    edge_port_enable_lacp: bool
    #: The type of encapsulation used on this edge port, by default DOT1Q.
    edge_port_encapsulation: EncapsulationType = EncapsulationType.DOT1Q
    #: The MAC address assigned to this edge port, if applicable.
    edge_port_mac_address: str | None = None
    #: The speed capacity of each member in the physical port.
    edge_port_member_speed: PhysicalPortCapacity
    #: The minimum number of links required for this edge port.
    edge_port_minimum_links: int | None = None
    #: The type of edge port (e.g., customer, private, public).
    edge_port_type: EdgePortType
    #: If set to True, the edge port will be ignored if it is down.
    edge_port_ignore_if_down: bool = False
    #: The GEANT GA ID associated with this edge port, if any.
    edge_port_geant_ga_id: str | None = None
    #: A list of LAG members associated with this edge port.
    edge_port_ae_members: LAGMemberList[LAGMember]
