"""Service Binding Port.

A service binding port is used to logically attach an edge port to a customer service using a :term:`VLAN`.
"""

from ipaddress import IPv4Address, IPv6Address
from typing import Annotated

from orchestrator.domain.base import ProductModel
from orchestrator.types import SubscriptionLifecycle
from pydantic import Field
from pydantic_forms.types import strEnum

VLANID = Annotated[int, Field(gt=0, lt=4096)]


class SBPType(strEnum):
    """Enumerator for the two allowed types of service binding port: layer 2 or layer 3."""

    L2 = "l2"
    L3 = "l3"


class ServiceBindingPortInactive(
    ProductModel, lifecycle=[SubscriptionLifecycle.INITIAL], product_block_name="ServiceBindingPort"
):
    """A Service Binding Port that's currently inactive. See :class:`ServiceBindingPort`."""

    is_tagged: bool | None = None
    vlan_id: VLANID | None = None
    sbp_type: SBPType | None = None
    ipv4_address: IPv4Address | None = None
    ipv6_address: IPv6Address | None = None
    custom_firewall_filters: bool | None = None
    geant_sid: str | None = None


class ServiceBindingPortProvisioning(ServiceBindingPortInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """A Service Binding Port that's currently being provisioned. See :class:`ServiceBindingPort`."""

    is_tagged: bool
    vlan_id: VLANID | None = None
    sbp_type: SBPType
    ipv4_address: IPv4Address | None = None
    ipv6_address: IPv6Address | None = None
    custom_firewall_filters: bool
    geant_sid: str


class ServiceBindingPort(ServiceBindingPortProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """A service binding port that is actively used in the network."""

    #: Whether this :term:`VLAN` is tagged or not.
    is_tagged: bool
    #: The :term:`VLAN` ID.
    vlan_id: VLANID | None = None
    #: Is this service binding port layer 2 or 3?
    sbp_type: SBPType
    #: If layer 3, IPv4 resources.
    ipv4_address: IPv4Address | None = None
    #: If layer 3, IPv6 resources.
    ipv6_address: IPv6Address | None = None
    #: Any custom firewall filters that the partner may require.
    custom_firewall_filters: bool
    #: The GÉANT service ID of this binding port.
    geant_sid: str
