"""A creation workflow for adding a new edge port to the network."""

from typing import Annotated, Any, Self

from annotated_types import Len
from orchestrator import step, workflow
from orchestrator.forms import FormPage
from orchestrator.forms.validators import Choice
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle, UUIDstr
from orchestrator.utils.errors import ProcessFailureError
from orchestrator.workflow import StepList, begin, done
from orchestrator.workflows.steps import resync, set_status, store_process_subscription
from orchestrator.workflows.utils import wrap_create_initial_input_form
from pydantic import AfterValidator, ConfigDict, model_validator
from pydantic_forms.validators import validate_unique_list
from pynetbox.models.dcim import Interfaces

from gso.products.product_blocks.edge_port import EdgePortType, EncapsulationType
from gso.products.product_blocks.router import RouterRole
from gso.products.product_types.edge_port import EdgePortInactive, EdgePortProvisioning
from gso.products.product_types.router import Router
from gso.services import lso_client, subscriptions
from gso.services.lso_client import lso_interaction
from gso.services.netbox_client import NetboxClient
from gso.utils.helpers import (
    available_interfaces_choices,
    available_service_lags_choices,
    partner_choice,
    validate_edge_port_number_of_members_based_on_lacp,
)
from gso.utils.types.interfaces import LAGMember, PhysicalPortCapacity
from gso.utils.types.tt_number import TTNumber


def initial_input_form_generator(product_name: str) -> FormGenerator:
    """Gather information to create a new Edge Port."""
    routers = {}
    for router in subscriptions.get_active_subscriptions_by_field_and_value("router_role", RouterRole.PE):
        routers[str(router.subscription_id)] = router.description
    router_enum = Choice("Select a router", zip(routers.keys(), routers.items(), strict=True))  # type: ignore[arg-type]

    class CreateEdgePortForm(FormPage):
        model_config = ConfigDict(title=product_name)

        tt_number: TTNumber
        node: router_enum  # type: ignore[valid-type]
        partner: partner_choice()  # type: ignore[valid-type]
        service_type: EdgePortType
        enable_lacp: bool
        speed: PhysicalPortCapacity
        encapsulation: EncapsulationType = EncapsulationType.DOT1Q
        number_of_members: int
        minimum_links: int
        mac_address: str | None = None
        ignore_if_down: bool = False
        geant_ga_id: str | None = None

        @model_validator(mode="after")
        def validate_number_of_members(self) -> Self:
            validate_edge_port_number_of_members_based_on_lacp(
                enable_lacp=self.enable_lacp, number_of_members=self.number_of_members
            )
            return self

    initial_user_input = yield CreateEdgePortForm

    class EdgePortLAGMember(LAGMember):
        interface_name: available_interfaces_choices(  # type: ignore[valid-type]
            initial_user_input.node, initial_user_input.speed
        )

    lag_ae_members = Annotated[
        list[EdgePortLAGMember],
        AfterValidator(validate_unique_list),
        Len(
            min_length=initial_user_input.number_of_members,
            max_length=initial_user_input.number_of_members,
        ),
    ]

    class SelectInterfaceForm(FormPage):
        model_config = ConfigDict(title="Select Interfaces")

        name: available_service_lags_choices(initial_user_input.node)  # type: ignore[valid-type]
        description: str | None = None
        ae_members: lag_ae_members

    interface_form_input_data = yield SelectInterfaceForm
    return initial_user_input.model_dump() | interface_form_input_data.model_dump()


@step("Create subscription")
def create_subscription(product: UUIDstr, partner: UUIDstr) -> State:
    """Create a new subscription object."""
    subscription = EdgePortInactive.from_product_id(product, partner)

    return {
        "subscription": subscription,
        "subscription_id": subscription.subscription_id,
    }


@step("Initialize subscription")
def initialize_subscription(
    subscription: EdgePortInactive,
    node: UUIDstr,
    service_type: EdgePortType,
    speed: PhysicalPortCapacity,
    encapsulation: EncapsulationType,
    name: str,
    minimum_links: int,
    geant_ga_id: str | None,
    mac_address: str | None,
    partner: str,
    enable_lacp: bool,  # noqa: FBT001
    ignore_if_down: bool,  # noqa: FBT001
    ae_members: list[dict[str, Any]],
    description: str | None = None,
) -> State:
    """Initialise the subscription object in the service database."""
    router = Router.from_subscription(node).router
    subscription.edge_port.edge_port_node = router
    subscription.edge_port.edge_port_type = service_type
    subscription.edge_port.edge_port_enable_lacp = enable_lacp
    subscription.edge_port.edge_port_member_speed = speed
    subscription.edge_port.edge_port_encapsulation = encapsulation
    subscription.edge_port.edge_port_name = name
    subscription.edge_port.edge_port_minimum_links = minimum_links
    subscription.edge_port.edge_port_ignore_if_down = ignore_if_down
    subscription.edge_port.edge_port_geant_ga_id = geant_ga_id
    subscription.edge_port.edge_port_mac_address = mac_address
    subscription.description = f"Edge Port {name} on {router.router_fqdn}, {partner}, {geant_ga_id or ""}"
    subscription.edge_port.edge_port_description = description
    for member in ae_members:
        subscription.edge_port.edge_port_ae_members.append(LAGMember(**member))
    subscription = EdgePortProvisioning.from_other_lifecycle(subscription, SubscriptionLifecycle.PROVISIONING)

    return {"subscription": subscription}


@step("Reserve interfaces in NetBox")
def reserve_interfaces_in_netbox(subscription: EdgePortProvisioning) -> State:
    """Create the :term:`LAG` interfaces in NetBox and attach the lag interfaces to the physical interfaces."""
    nbclient = NetboxClient()
    edge_port = subscription.edge_port
    # Create :term:`LAG` interfaces
    lag_interface: Interfaces = nbclient.create_interface(
        iface_name=edge_port.edge_port_name,
        interface_type="lag",
        device_name=edge_port.edge_port_node.router_fqdn,
        description=str(subscription.subscription_id),
        enabled=True,
    )
    # Attach physical interfaces to :term:`LAG`
    # Update interface description to subscription ID
    # Reserve interfaces
    for interface in edge_port.edge_port_ae_members:
        nbclient.attach_interface_to_lag(
            device_name=edge_port.edge_port_node.router_fqdn,
            lag_name=lag_interface.name,
            iface_name=interface.interface_name,
            description=str(subscription.subscription_id),
        )
        nbclient.reserve_interface(
            device_name=edge_port.edge_port_node.router_fqdn,
            iface_name=interface.interface_name,
        )
    return {
        "subscription": subscription,
    }


@step("Allocate interfaces in NetBox")
def allocate_interfaces_in_netbox(subscription: EdgePortProvisioning) -> None:
    """Allocate the interfaces in NetBox."""
    for interface in subscription.edge_port.edge_port_ae_members:
        fqdn = subscription.edge_port.edge_port_node.router_fqdn
        iface_name = interface.interface_name
        if not fqdn or not iface_name:
            msg = "FQDN and/or interface name missing in subscription"
            raise ProcessFailureError(msg, details=subscription.subscription_id)

        NetboxClient().allocate_interface(device_name=fqdn, iface_name=iface_name)


@step("[DRY RUN] Create edge port")
def create_edge_port_dry(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Create a new edge port in the network as a dry run."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Create Edge Port",
        "verb": "create",
    }

    lso_client.execute_playbook(
        playbook_name="edge_port.yaml",
        callback_route=callback_route,
        inventory=subscription["edge_port"]["edge_port_node"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@step("[FOR REAL] Create edge port")
def create_edge_port_real(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Create a new edge port in the network for real."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Create Edge Port",
        "verb": "create",
    }

    lso_client.execute_playbook(
        playbook_name="edge_port.yaml",
        callback_route=callback_route,
        inventory=subscription["edge_port"]["edge_port_node"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@workflow(
    "Create Edge Port",
    initial_input_form=wrap_create_initial_input_form(initial_input_form_generator),
    target=Target.CREATE,
)
def create_edge_port() -> StepList:
    """Create a new edge port in the network.

    * Create and initialise the subscription object in the service database
    * Deploy configuration on the new edge port, first as a dry run
    * allocate LAG and LAG members in the Netbox.
    """
    return (
        begin
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> reserve_interfaces_in_netbox
        >> lso_interaction(create_edge_port_dry)
        >> lso_interaction(create_edge_port_real)
        >> allocate_interfaces_in_netbox
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
